@@ -34,7 +34,7 @@ __all__ = [
     'ImproperlyConfigured', 'SqliteDatabase', 'MySQLDatabase', 'PostgresqlDatabase',
     'asc', 'desc', 'Count', 'Max', 'Min', 'Q', 'Field', 'CharField', 'TextField',
     'DateTimeField', 'BooleanField', 'FloatField', 'IntegerField', 'PrimaryKeyField',
-    'ForeignKeyField', 'Model',
+    'ForeignKeyField', 'Model', 'filter_query', 'annotate_query',
 ]
 
 class ImproperlyConfigured(Exception):
@@ -123,6 +123,8 @@ class SqliteAdapter(BaseAdapter):
     interpolation = '?'
     
     def connect(self, database, **kwargs):
+        if not sqlite3:
+            raise ImproperlyConfigured('sqlite3 must be installed on the system')
         return sqlite3.connect(database, **kwargs)
     
     def lookup_cast(self, lookup, value):
@@ -154,6 +156,8 @@ class PostgresqlAdapter(BaseAdapter):
     }
         
     def connect(self, database, **kwargs):
+        if not psycopg2:
+            raise ImproperlyConfigured('psycopg2 must be installed on the system')
         return psycopg2.connect(database=database, **kwargs)
     
     def get_field_overrides(self):
@@ -185,6 +189,8 @@ class MySQLAdapter(BaseAdapter):
     }
 
     def connect(self, database, **kwargs):
+        if not mysql:
+            raise ImproperlyConfigured('MySQLdb must be installed on the system')
         return mysql.connect(db=database, **kwargs)
 
     def get_field_overrides(self):
@@ -548,7 +554,20 @@ class BaseQuery(object):
         
         self._dirty = True
         self._where = {}
-        self._joins = []
+        self._joins = {}
+        self._joined_models = set()
+    
+    def _clone_dict_graph(self, dg):
+        cloned = {}
+        for node, edges in dg.items():
+            cloned[node] = list(edges)
+        return cloned
+    
+    def clone_where(self):
+        return self._clone_dict_graph(self._where)
+    
+    def clone_joins(self):
+        return self._clone_dict_graph(self._joins)
     
     def clone(self):
         raise NotImplementedError
@@ -604,7 +623,9 @@ class BaseQuery(object):
     @returns_clone
     def join(self, model, join_type=None, on=None):
         if self.query_context._meta.rel_exists(model):
-            self._joins.append((model, join_type, on))
+            self._joined_models.add(model)
+            self._joins.setdefault(self.query_context, [])
+            self._joins[self.query_context].append((model, join_type, on))
             self.query_context = model
         else:
             raise AttributeError('No foreign key found between %s and %s' % \
@@ -616,74 +637,87 @@ class BaseQuery(object):
             self.query_context = model
             return
 
-        for klass, join_type, on in self._joins:
-            if model == klass:
-                self.query_context = model
-                return
+        if model in self._joined_models:
+            self.query_context = model
+            return
         raise AttributeError('You must JOIN on %s' % model.__name__)
     
     def use_aliases(self):
-        return len(self._joins) > 0 or self.force_alias
+        return len(self._joined_models) > 0 or self.force_alias
 
     def combine_field(self, alias, field_name):
         if alias:
             return '%s.%s' % (alias, field_name)
         return field_name
     
+    def follow_joins(self, current, alias_map, alias_required, alias_count, seen=None):
+        computed = []
+        seen = seen or set()
+        
+        if current not in self._joins:
+            return computed
+        
+        for i, (model, join_type, on) in enumerate(self._joins[current]):
+            seen.add(model)
+            
+            if alias_required:
+                alias_count += 1
+                alias_map[model] = 't%d' % alias_count
+            else:
+                alias_map[model] = ''
+            
+            from_model = current
+            field = from_model._meta.get_related_field_for_model(model, on)
+            if field:
+                left_field = field.name
+                right_field = model._meta.pk_name
+            else:
+                field = from_model._meta.get_reverse_related_field_for_model(model, on)
+                left_field = from_model._meta.pk_name
+                right_field = field.name
+            
+            if join_type is None:
+                if field.null and model not in self._where:
+                    join_type = 'LEFT OUTER'
+                else:
+                    join_type = 'INNER'
+            
+            computed.append(
+                '%s JOIN %s AS %s ON %s = %s' % (
+                    join_type,
+                    model._meta.db_table,
+                    alias_map[model],
+                    self.combine_field(alias_map[from_model], left_field),
+                    self.combine_field(alias_map[model], right_field),
+                )
+            )
+            
+            computed.extend(self.follow_joins(model, alias_map, alias_required, alias_count, seen))
+        
+        return computed
+    
     def compile_where(self):
         alias_count = 0
         alias_map = {}
 
         alias_required = self.use_aliases()
 
-        joins = list(self._joins)
-        if self._where or len(joins):
-            joins.insert(0, (self.model, None, None))
-        
         where_with_alias = []
         where_data = []
-        computed_joins = []
-
-        for i, (model, join_type, on) in enumerate(joins):
-            if alias_required:
-                alias_count += 1
-                alias_map[model] = 't%d' % alias_count
-            else:
-                alias_map[model] = ''
-            
-            if i > 0:
-                from_model = joins[i-1][0]
-                field = from_model._meta.get_related_field_for_model(model, on)
-                if field:
-                    left_field = field.name
-                    right_field = model._meta.pk_name
-                else:
-                    field = from_model._meta.get_reverse_related_field_for_model(model, on)
-                    left_field = from_model._meta.pk_name
-                    right_field = field.name
-                
-                if join_type is None:
-                    if field.null and model not in self._where:
-                        join_type = 'LEFT OUTER'
-                    else:
-                        join_type = 'INNER'
-                
-                computed_joins.append(
-                    '%s JOIN %s AS %s ON %s = %s' % (
-                        join_type,
-                        model._meta.db_table,
-                        alias_map[model],
-                        self.combine_field(alias_map[from_model], left_field),
-                        self.combine_field(alias_map[model], right_field),
-                    )
-                )
         
-        for (model, join_type, on) in joins:
-            if model in self._where:
-                for node in self._where[model]:
-                    query, data = self.parse_node(node, model, alias_map)
-                    where_with_alias.append(query)
-                    where_data.extend(data)
+        if alias_required:
+            alias_count += 1
+            alias_map[self.model] = 't%d' % alias_count
+        else:
+            alias_map[self.model] = ''
+        
+        computed_joins = self.follow_joins(self.model, alias_map, alias_required, alias_count)
+        
+        for model in sorted(self._where, key=lambda m: alias_map[m]):
+            for node in self._where[model]:
+                query, data = self.parse_node(node, model, alias_map)
+                where_with_alias.append(query)
+                where_data.extend(data)
         
         return computed_joins, where_with_alias, where_data, alias_map
     
@@ -801,8 +835,9 @@ class SelectQuery(BaseQuery):
         query._pagination = self._pagination and tuple(self._pagination) or None
         query._distinct = self._distinct
         query._qr = self._qr
-        query._where = dict(self._where)
-        query._joins = list(self._joins)
+        query._where = self.clone_where()
+        query._joined_models = self._joined_models.copy()
+        query._joins = self.clone_joins()
         return query
     
     @returns_clone
@@ -878,16 +913,27 @@ class SelectQuery(BaseQuery):
             ))
         finally:
             self.query_context = orig_ctx
+    
+    def filter(self, *args, **kwargs):
+        return filter_query(self, *args, **kwargs)
+    
+    def annotate(self, related_model, aggregation=None):
+        return annotate_query(self, related_model, aggregation)
 
     def parse_select_query(self, alias_map):
-        if isinstance(self.query, basestring):
-            if self.query in ('*', self.model._meta.pk_name) and self.use_aliases():
-                return '%s.%s' % (alias_map[self.model], self.query)
-            return self.query
-        elif isinstance(self.query, dict):
+        if isinstance(self.query, (list, tuple)):
+            query = {self.model: self.query}
+        else:
+            query = self.query
+        
+        if isinstance(query, basestring):
+            if query in ('*', self.model._meta.pk_name) and self.use_aliases():
+                return '%s.%s' % (alias_map[self.model], query)
+            return query
+        elif isinstance(query, dict):
             qparts = []
             aggregates = []
-            for model, cols in self.query.iteritems():
+            for model, cols in query.iteritems():
                 alias = alias_map.get(model, '')
                 for col in cols:
                     if isinstance(col, tuple):
@@ -983,8 +1029,9 @@ class UpdateQuery(BaseQuery):
     
     def clone(self):
         query = UpdateQuery(self.model, **self.update_query)
-        query._where = dict(self._where)
-        query._joins = list(self._joins)
+        query._where = self.clone_where()
+        query._joined_models = self._joined_models.copy()
+        query._joins = self.clone_joins()
         return query
     
     def parse_update(self):
@@ -1035,8 +1082,9 @@ class UpdateQuery(BaseQuery):
 class DeleteQuery(BaseQuery):
     def clone(self):
         query = DeleteQuery(self.model)
-        query._where = dict(self._where)
-        query._joins = list(self._joins)
+        query._where = self.clone_where()
+        query._joined_models = self._joined_models.copy()
+        query._joins = self.clone_joins()
         return query
     
     def sql(self):
@@ -1100,6 +1148,12 @@ class InsertQuery(BaseQuery):
         return self.database.last_insert_id(result, self.model)
 
 
+def model_or_select(m_or_q):
+    if isinstance(m_or_q, BaseQuery):
+        return (m_or_q.model, m_or_q)
+    else:
+        return (m_or_q, m_or_q.select())
+
 def convert_lookup(model, joins, lookup):
     operations = model._meta.database.adapter.operations
     
@@ -1147,10 +1201,12 @@ def convert_lookup(model, joins, lookup):
     return query_model, joins, lookup
 
 
-def filter_query(model, *args, **kwargs):
+def filter_query(model_or_query, *args, **kwargs):
     """
     Provide a django-like interface for executing queries
     """
+    model, select_query = model_or_select(model_or_query)
+    
     query = {} # mapping of models to queries
     joins = {} # a graph of joins needed, passed into the convert_lookup function
     
@@ -1184,10 +1240,12 @@ def filter_query(model, *args, **kwargs):
     def follow_joins(current, query):
         if current in joins:
             for joined_model in joins[current]:
-                query = query.join(joined_model)
+                query = query.switch(current)
+                if joined_model not in query._joined_models:
+                    query = query.join(joined_model)
                 query = follow_joins(joined_model, query)
         return query
-    select_query = follow_joins(model, model.select())
+    select_query = follow_joins(model, select_query)
     
     for model, lookups in query.items():
         qargs, qkwargs = [], {}
@@ -1200,6 +1258,39 @@ def filter_query(model, *args, **kwargs):
 
     return select_query
 
+def annotate_query(select_query, related_model, aggregation):
+    aggregation = aggregation or Count(related_model._meta.pk_name)
+    model = select_query.model
+    
+    select_query = select_query.switch(model)
+    cols = select_query.query
+    
+    # ensure the join is there
+    if related_model not in select_query._joined_models:
+        select_query = select_query.join(related_model).switch(model)
+    
+    # query for it
+    if isinstance(cols, dict):
+        selection = cols
+        group_by = cols[model]
+    elif isinstance(cols, basestring):
+        selection = {model: [cols]}
+        if cols == '*':
+            group_by = model
+        else:
+            group_by = [col.strip() for col in cols.split(',')]
+    elif isinstance(cols, (list, tuple)):
+        selection = {model: cols}
+        group_by = cols
+    else:
+        raise ValueError('Unknown type passed in to select query: "%s"' % type(cols))
+    
+    # query for the related object
+    selection[related_model] = [aggregation]
+    
+    select_query.query = selection
+    return select_query.group_by(group_by)
+
 
 class Field(object):
     db_field = ''
